"""
Copyright 2019, University of Freiburg.

Niklas Schnelle <schnelle@cs.uni-freiburg.de>
"""
from typing import List, Optional, Dict, Any, Set, DefaultDict, NamedTuple, Tuple as T  # noqa: F401, E501 pylint: disable=unused-import
from abc import ABC, abstractmethod


class BackendError(Exception):
    """
    An unspecified backend errror
    TODO(schnelle) better errors
    """


class Backend(ABC):
    """
    Backend abstracts a SPARQL endpoint
    """
    def __init__(self):
        self.last_query_client_time = None  # type: Optional[float]
        self.last_query_compute_time = None  # type: Optional[float]
        self.last_query_time = None  # type: Optional[float]
        self.is_qlever = False

    @abstractmethod
    def query(self, query: str, **kwargs) \
            -> T[Optional[List[List[str]]], Dict[str, Any]]:
        """Returns the result table of the query as a list of rows.

        :param query:
        :return:
        """
