from query_model import query_model
from copy import deepcopy


def create_plus_path_from_primary(primary: str, mod: str):
    """
    input is a single iri <i>, output is the property path
    <i> +

    """
    if mod == "*":
        mod = "+"
    elif mod == "?":
        mod = None
    else:
        raise ValueError("Invalid Modifier when hacking QLever's property path optimizer. Please report this")
    elt = query_model.PropertyPath.PathElt(primary, False, mod)
    seq = query_model.PropertyPath.PathSequence([elt])
    return query_model.PropertyPath([seq])

def qlever_property_path_hack(query : query_model.SelectQuery):
    """
    Takes a query. If that query contains at least one propery path
    of the form <a>/<b>*, the first of such paths will be rewritten
    by an equivalent union that can currently be dealt with more efficiently
    by QLever in most cases.
    :return:
    """
    out = where = new_where = None
    if "ql:has-predicate" in query.to_sparql():
        return query
    if isinstance(query, query_model.SelectQuery):
        out = deepcopy(query)
        where = out.where
        new_where = query_model.GroupGraphPattern()
        out.where = new_where
    elif isinstance(query, query_model.GroupGraphPattern):
        # in this case we return a new group graph pattern
        where = deepcopy(query)
        new_where = query_model.GroupGraphPattern()
        out = new_where
    else:
        raise TypeError("Unsupported input type to qlever_property_path_hack")

    current_context = query_model.GroupGraphPattern()
    specialPattern = None
    specialPatternPos = -1
    numNewVariables = 0
    literalFilter = None
    for pattern in where:
        if isinstance(pattern, list):
            if not isinstance(pattern[1], query_model.PropertyPath):
                current_context.append(pattern)
                continue
            try:
              fst, snd, modifier = pattern[1].getSpecialPath()
              if specialPattern is None and str(pattern[0]).startswith("?") and str(pattern[2]).startswith("?"):
                  specialPattern = (pattern[0], (fst, create_plus_path_from_primary(snd, modifier)), pattern[2])
                  specialPatternPos = len(current_context)
              else:
                  # only greedily replace the first property path for now
                  current_context.append(pattern)
            except query_model.PropertyPath.NotSpeciallError:
                current_context.append(pattern)
                pass
        elif isinstance(pattern, query_model.OptionalPattern):
            opt_context = qlever_property_path_hack(pattern.pattern)
            opt = query_model.OptionalPattern(opt_context)
            opt.parent = new_where
            current_context.append(opt)
            continue
        elif isinstance(pattern, query_model.MinusGraphPattern):
            minus_context = qlever_property_path_hack(pattern.pattern)
            minus = query_model.MinusGraphPattern(minus_context)
            minus.parent = new_where
            current_context.append(minus)
            continue
        elif isinstance(pattern, query_model.UnionPattern):
            lhs_context = qlever_property_path_hack(pattern.lhs)
            rhs_context = qlever_property_path_hack(pattern.rhs)
            union = query_model.UnionPattern(lhs_context, rhs_context)
            union.parent = new_where
            current_context.append(union)
            continue
        elif isinstance(pattern, query_model.GroupGraphPattern):
            current_context.append(qlever_property_path_hack(pattern))
            continue
        elif isinstance(pattern, query_model.SelectQuery):
            # recurse on subqueries
            current_context.append(qlever_property_path_hack(pattern))
        elif isinstance(pattern, query_model.Filter) :
            current_context.append(pattern)

            if ("\"^\\\"" in pattern.text or "=\"" in pattern.text) and "?ql_entity" in pattern.variables:
                literalFilter = pattern
        else :
            #any other patterns
            current_context.append(pattern)
    if len(current_context) or specialPattern is not None :
        if specialPattern is None:
            new_where.extend(current_context)
        else:
            l = deepcopy(current_context)
            r = deepcopy(current_context)
            l.insert(specialPatternPos, [specialPattern[0], specialPattern[1][0], specialPattern[2]])

            new_variable = "?qlever_hacks_new_variable" + str(numNewVariables)
            numNewVariables += 1

            r.insert(specialPatternPos, [specialPattern[0], specialPattern[1][0], new_variable])
            #if str(specialPattern[1][1]) == "wdt:P279+" and str(specialPattern[1][0]) != "wdt:P31":
            if False: # this only applies when we have wdt:P279+ in the cache which we currently don't
                w = query_model.GroupGraphPattern([[new_variable, specialPattern[1][1], specialPattern[2]]])
                s = query_model.SelectQuery(query_model.Prologue("", []), True, [new_variable, specialPattern[2]], w, None,
                                            None, None)
                s.parent = r
                r.insert(specialPatternPos + 1, s)
            else:
                r.insert(specialPatternPos + 1, [new_variable, specialPattern[1][1], specialPattern[2]])

            if literalFilter is not None:
                innerWhere = query_model.GroupGraphPattern()
                innerWhere.append(["?tmp", specialPattern[1][1].remove_modifiers(), specialPattern[2]])
                innerWhere.append(literalFilter)
                innerSubquery = query_model.SelectQuery(query_model.Prologue("", []), True, [specialPattern[2]],
                                            innerWhere, None, None, None)
                innerSubquery.parent = r
                r.insert(specialPatternPos + 2, innerSubquery)

            u = query_model.UnionPattern(l, r)
            g = query_model.GroupGraphPattern()
            g.append(u)
            selected = [el for el in u.all_variables() if not el.startswith("?qlever_hacks_new_variable")]
            s = query_model.SelectQuery(query_model.Prologue("", []), True, selected,
                                        g, None, None, None)
            new_where.append(s)
    return out




