

import re

def substitute(input, templates, require_all=False):
    previous_pos = 0
    output = ""
    pos = input.find("%")
    while pos != -1:
        output += input[previous_pos:pos]
        next = input.find("%", pos + 1)
        if next == -1 :
            raise ValueError("Mismatching number of Percent Signs % ... % in input query template")
        template_name = input[pos + 1 : next]
        if template_name not in templates and require_all:
            raise ValueError("No template named \"" + template_name + "\" found for substitution")
        if template_name in templates:
            output += substitute(templates[template_name], templates)
        else:
            output += input[pos:next+1]
        previous_pos = next + 1
        pos = input.find("%", previous_pos)
    output += input[previous_pos:]
    return output

def evaluate_conditions(input, conditions, require_all=False):
    if (require_all and not "PREFIX_COMPLETE" in conditions):
        conditions["PREFIX_COMPLETE"] = False
    input_orig = input
    for condition in conditions :
        positive = "TRUE" if conditions[condition] else "FALSE"
        negative = "FALSE" if conditions[condition] else "TRUE"
        input = re.sub("# IF " + condition + " #", "# IF {} #".format(positive), input)
        input = re.sub("# IF !" + condition + " #", "# IF {} #".format(negative), input)

    input_save = input
    while True:
        # this will substitute the innermost level of ifs first, repeat until we can't anymore
        re_simple_true = "# IF TRUE #([^#]*)# ENDIF #"
        input = re.sub(re_simple_true, r"\1", input)
        re_simple_false = "# IF FALSE #([^#]*)# ENDIF #"
        input = re.sub(re_simple_false, "", input)
        re_else_true = "# IF TRUE #([^#]*)# ELSE #([^#]*)# ENDIF #"
        input = re.sub(re_else_true, r"\1", input)
        re_else_false = "# IF FALSE #([^#]*)# ELSE #([^#]*)# ENDIF #"
        input = re.sub(re_else_false, r"\2", input)
        if input == input_save:
            break
        input_save = input

    if require_all and ("# ENDIF #" in input or "# IF " in input ):
        raise ValueError("Could not completely evaluate the conditions in template \n{} with conditions\n{}. Result that still contains \'#\' is\n{}".format(input_orig, conditions, input))
    return input


if __name__ == "__main__":
    template = """
    PREFIX rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#>
PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>
PREFIX skos: <http://www.w3.org/2004/02/skos/core#>
PREFIX schema: <http://schema.org/>
PREFIX wikibase: <http://wikiba.se/ontology#>
PREFIX wdt: <http://www.wikidata.org/prop/direct/>
SELECT ?qleverui_entity (SAMPLE(?name) AS ?qleverui_name) (SAMPLE(?alias) AS ?qleverui_altname) (SAMPLE(?count) AS ?qleverui_count) WHERE {
  { SELECT ?qleverui_entity ?name ?alias ?count WHERE {
    ?qleverui_entity rdfs:label ?name .
    ?qleverui_entity rdfs:label|@en@skos:altLabel ?alias .
    ?qleverui_entity ^schema:about/wikibase:sitelinks ?count .
  } ORDER BY ?alias }
  # IF CURRENT_WORD_EMPTY #
  ?qleverui_entity wdt:P31 wd:Q1 .
  # ENDIF #
  # IF !CURRENT_WORD_EMPTY #
  FILTER regex(?alias, "^\"%CURRENT_WORD%")
  # ENDIF #
} GROUP BY ?qleverui_entity ORDER BY DESC(?qleverui_count)
    """
    evaluate_conditions(template, {"CURRENT_WORD_EMPTY": True})











